#include <stdint.h>

#include <at45.h>
#include <config.h>
#include <serial.h>
#include <spi.h>
#include <utils.h>

#define AT45		"AT45: "

static int at45_cs;

static void
at45_transfer(	void *cmd_ptr,
		int cmd_len,
		void *dat_ptr,
		int dat_len)
{
	struct spi_tblk blks[2];
	int blk_n = 1;

	blks[0].rx = cmd_ptr;
	blks[0].tx = cmd_ptr;
	blks[0].len = cmd_len;

	if (dat_len != 0) {
		blks[1].rx = dat_ptr;
		blks[1].tx = dat_ptr;
		blks[1].len = dat_len;
		++blk_n;
	}

	spi_select(at45_cs);
	spi_transfer(blks, blk_n);
}

static int
at45_status()
{
	uint8_t req[2];

	req[0] = 0xd7;
	at45_transfer(req, 2, 0, 0);

	return req[1];
}

static void
at45_wait_rdy()
{
	while (!(at45_status() & 0x80)) ;
}

static void
at45_buffer_fill(int id, const void *in)
{
	uint8_t buffer[AT45_PAGE_SIZE];
	uint8_t req[4];

	req[0] = (id == 1) ? 0x84 : 0x87;
	req[2] = 0;
	req[3] = 0;

	memcopy(buffer, in, AT45_PAGE_SIZE);

	at45_transfer(req, 4, buffer, AT45_PAGE_SIZE);
}

static void
at45_page_write(int id, uint8_t page)
{
	uint8_t req[4];

	req[0] = (id == 1) ? 0x83 : 0x86;
	req[1] = ((page & 0x0000ff00) >> 7);
	req[2] = ((page & 0x000000ff) << 1);

	at45_transfer(req, 4, 0, 0);
}

int at45_init(int cs)
{
	uint8_t req[4];
	at45_cs = cs;

	spi_mode(at45_cs, SPI_MODE_0, 8, 5, NULL);

	req[0] = 0x9f;
	at45_transfer(req, 4, 0, 0);

	serial_printk(AT45 "Manufacturer ID: %b\n", req[1]);
	serial_printk(AT45 "Density Code: %b\n", req[2]);
	serial_printk(AT45 "Product Version: %b\n", req[3]);

	if (req[1] != 0x1f || req[2] != 0x25) {
		serial_printk(AT45 "Unknown Device\n");
		return -1;
	}

	serial_printk(AT45 "Status Register: %b\n", at45_status());

	return 0;
}

int at45_read(uint32_t lba, void *ptr, int len)
{
	uint8_t req[5];
	uint32_t addr, page, byte;

	__u32divmod(lba, AT45_PAGE_SIZE, &page, &byte);
	addr = (page << 9) | byte;

	serial_printk(AT45 "Reading addr = %i, length = %i\n",
			addr, len);

	req[0] = 0x0b;
	req[1] = ((addr & 0x00ff0000) >> 16);
	req[2] = ((addr & 0x0000ff00) >> 8);
	req[3] = ((addr & 0x000000ff));

	at45_transfer(req, 5, ptr, len);

	return 0;
}

/* FIXME: It should be <len> == AT45_PAGE_SIZE * <n>
 * */
int at45_write(uint32_t lba, const void *ptr, int len)
{
	const uint8_t *ptr8_r = ptr;
	uint32_t page, pg_n;
	uint32_t z0, z1;
	int id = 1;

	__u32divmod(lba, AT45_PAGE_SIZE, &page, &z0);
	__u32divmod(len, AT45_PAGE_SIZE, &pg_n, &z1);

	if (z0 != 0)
		return -1;

	if (z1 != 0)
		return -1;

	serial_printk(AT45 "Writing page = %i, count = %i\n",
			page, pg_n);

	while (pg_n--) {
		at45_buffer_fill(id, ptr8_r);

		at45_wait_rdy();
		at45_page_write(id, page++);

		id = (id == 1) ? 2 : 1;
		ptr8_r += AT45_PAGE_SIZE;
	}

	at45_wait_rdy();

	return 0;
}

