#include <stdint.h>

#include <ext2.h>
#include <mbr.h>
#include <mci_sd.h>
#include <config.h>
#include <sdram.h>
#include <serial.h>
#include <utils.h>

#define EXT2				"EXT2: "

#define EXT2_SUPER_MAGIC		0xEF53

/*
 * Special inode numbers
 */
#define	EXT2_BAD_INO			 1	/* Bad blocks inode */
#define EXT2_ROOT_INO			 2	/* Root inode */
#define EXT2_BOOT_LOADER_INO		 5	/* Boot loader inode */
#define EXT2_UNDEL_DIR_INO		 6	/* Undelete directory inode */

/*
 * Constants relative to the data blocks
 */
#define	EXT2_NDIR_BLOCKS		12
#define	EXT2_IND_BLOCK			EXT2_NDIR_BLOCKS
#define	EXT2_DIND_BLOCK			(EXT2_IND_BLOCK + 1)
#define	EXT2_TIND_BLOCK			(EXT2_DIND_BLOCK + 1)
#define	EXT2_N_BLOCKS			(EXT2_TIND_BLOCK + 1)

#define EXT2_NAME_LEN 			255

/*
 * Structure of the super block
 */
struct ext2_super_block {
	uint32_t	s_inodes_count;		/* Inodes count */
	uint32_t	s_blocks_count;		/* Blocks count */
	uint32_t	s_r_blocks_count;	/* Reserved blocks count */
	uint32_t	s_free_blocks_count;	/* Free blocks count */
	uint32_t	s_free_inodes_count;	/* Free inodes count */
	uint32_t	s_first_data_block;	/* First Data Block */
	uint32_t	s_log_block_size;	/* Block size */
	uint32_t	s_log_frag_size;	/* Fragment size */
	uint32_t	s_blocks_per_group;	/* # Blocks per group */
	uint32_t	s_frags_per_group;	/* # Fragments per group */
	uint32_t	s_inodes_per_group;	/* # Inodes per group */
	uint32_t	s_mtime;		/* Mount time */
	uint32_t	s_wtime;		/* Write time */
	uint16_t	s_mnt_count;		/* Mount count */
	uint16_t	s_max_mnt_count;	/* Maximal mount count */
	uint16_t	s_magic;		/* Magic signature */
	uint16_t	s_state;		/* File system state */
	uint16_t	s_errors;		/* Behaviour when detecting errors */
	uint16_t	s_minor_rev_level; 	/* minor revision level */
	uint32_t	s_lastcheck;		/* time of last check */
	uint32_t	s_checkinterval;	/* max. time between checks */
	uint32_t	s_creator_os;		/* OS */
	uint32_t	s_rev_level;		/* Revision level */
	uint16_t	s_def_resuid;		/* Default uid for reserved blocks */
	uint16_t	s_def_resgid;		/* Default gid for reserved blocks */
	uint32_t	s_first_ino; 		/* First non-reserved inode */
	uint16_t  	s_inode_size; 		/* size of inode structure */
	uint16_t	s_block_group_nr; 	/* block group # of this superblock */
	uint32_t	s_feature_compat; 	/* compatible feature set */
	uint32_t	s_feature_incompat; 	/* incompatible feature set */
	uint32_t	s_feature_ro_compat; 	/* readonly-compatible feature set */
	uint8_t		s_uuid[16];		/* 128-bit uuid for volume */
	char		s_volume_name[16]; 	/* volume name */
	char		s_last_mounted[64]; 	/* directory where last mounted */
	uint32_t	s_algorithm_usage_bitmap; /* For compression */
} __attribute__ ((__packed__));

/*
 * Structure of an inode on the disk
 */
struct ext2_inode {
	uint16_t	i_mode;			/* File mode */
	uint16_t	i_uid;			/* Low 16 bits of Owner Uid */
	uint32_t	i_size;			/* Size in bytes */
	uint32_t	i_atime;		/* Access time */
	uint32_t	i_ctime;		/* Creation time */
	uint32_t	i_mtime;		/* Modification time */
	uint32_t	i_dtime;		/* Deletion Time */
	uint16_t	i_gid;			/* Low 16 bits of Group Id */
	uint16_t	i_links_count;		/* Links count */
	uint32_t	i_blocks;		/* Blocks count */
	uint32_t	i_flags;		/* File flags */
	uint32_t	i_reserved1;
	uint32_t	i_block[EXT2_N_BLOCKS];	/* Pointers to blocks */
	uint32_t	i_generation;		/* File version (for NFS) */
	uint32_t	i_file_acl;		/* File ACL */
	uint32_t	i_dir_acl;		/* Directory ACL */
	uint32_t	i_faddr;		/* Fragment address */
	uint8_t		i_frag;			/* Fragment number */
	uint8_t		i_fsize;		/* Fragment size */
	uint16_t	i_pad1;
	uint16_t	i_uid_high;		/* these 2 fields    */
	uint16_t	i_gid_high;		/* were reserved2[0] */
	uint32_t	i_reserved2;
} __attribute__ ((__packed__));

/*
 * Structure of a blocks group descriptor
 */
struct ext2_group_desc {
	uint32_t	bg_block_bitmap;	/* Blocks bitmap block */
	uint32_t	bg_inode_bitmap;	/* Inodes bitmap block */
	uint32_t	bg_inode_table;		/* Inodes table block */
	uint16_t	bg_free_blocks_count;	/* Free blocks count */
	uint16_t	bg_free_inodes_count;	/* Free inodes count */
	uint16_t	bg_used_dirs_count;	/* Directories count */
	uint16_t	bg_pad;
	uint32_t	bg_reserved[3];
} __attribute__ ((__packed__));

/*
 * Structure of a directory entry
 */
struct ext2_dir_entry {
	uint32_t	inode;			/* Inode number */
	uint16_t	rec_len;		/* Directory entry length */
	uint16_t	name_len;		/* Name length */
	char	name[EXT2_NAME_LEN];		/* File name */
} __attribute__ ((__packed__));

/* Simple memory allocator (We have large SDRAM)
 * */
static uint8_t *ext2_mem_ptr;

static void *
ext2_mem_alloc(int len)
{
	void *ret = ext2_mem_ptr;
	int pad = (4 - (len & 3)) & 3;

	ext2_mem_ptr += len + pad;

	return ret;
}

struct ext2_mount_data {
	struct ext2_super_block super_block;
	struct ext2_inode root_inode;
	struct ext2_group_desc *gdesc_table;
	struct ext2_inode curr_inode;
	int (*lread) (uint32_t, int, void*);
	uint32_t block_size;			/* Ext2 block in bytes */
	uint32_t beg;				/* First block of boot partition */
	void *block;				/* Buffer */
	uint32_t *intab;			/* Indirect blocks table */
	uint32_t *dintab;			/* Double indirect blocks table */
};

static struct ext2_mount_data *ext2;

/* Search for boot partition
 * */
static int
ext2_get_partition(int num, uint32_t *ret)
{
	struct msdos_partition *par;
	uint8_t *mbr = ext2_mem_alloc(EXT2_DEVICE_BLKLEN);
	uint16_t *label;

	ext2->lread(0, 1, mbr);

	label = (uint16_t*) (mbr + 510);
	if (*label != MSDOS_LABEL_MAGIC) {
		serial_printk(EXT2 "%h Label incorrect, expected %h \n", *label, MSDOS_LABEL_MAGIC);
		return -1;
	}

	par = (struct msdos_partition *) (mbr + 446 + 16 * num);
	if (par->system_id != LINUX_SYSTEM_ID) {
		serial_printk(EXT2 "%b ID incorrect, expected %b \n", par->system_id, LINUX_SYSTEM_ID);
		return -1;
	}

	*ret = par->sect_beg;

	serial_printk(EXT2 "Partition [%i] offset: %i \n", num, *ret);

	return 0;
}

/* Read specified inode
 * */
static int
ext2_read_inode(uint32_t i_num, struct ext2_inode *inode)
{
	struct ext2_group_desc *gd;
	uint32_t grp, ind;
	uint32_t pos, blk, ofs;

 	__u32divmod((i_num - 1), ext2->super_block.s_inodes_per_group, &grp, &ind);

	gd = ext2->gdesc_table + grp;
	pos = gd->bg_inode_table * ext2->block_size +
		ind * ext2->super_block.s_inode_size;

	__u32divmod(pos, EXT2_DEVICE_BLKLEN, &blk, &ofs);

	ext2->lread(ext2->beg + blk, 1, ext2->block);
	memcopy(inode, (uint8_t*) ext2->block + ofs,
			sizeof(struct ext2_inode));

	return 0;
}

/* Search file in directory, and refresh inode
 * */
static int
ext2_dir_follow(const char *name, struct ext2_inode *inode)
{
	struct ext2_dir_entry *dir;
	int i, hit;

	//FIXME: it's not correct, all blocks should be readed,
	// not only first
	ext2->lread(ext2->beg + inode->i_block[0] *
			ext2->block_size / EXT2_DEVICE_BLKLEN,
			ext2->block_size / EXT2_DEVICE_BLKLEN, ext2->block);

	dir = (struct ext2_dir_entry *) ext2->block;

	while (dir->name_len) {

		hit = 1;
		for (i = 0; (i < dir->name_len) && (name[i]); ++i) {
			if (dir->name[i] != name[i]) {
				hit = 0;
				break;
			}
		}

		if (hit) {
			ext2_read_inode(dir->inode, inode);
			return 0;
		}

		dir = (struct ext2_dir_entry *) ((uint8_t*) dir + dir->rec_len);
	}

	return -1;
}

/* Block reading system, for optimize reading multiple blocks
 * */
struct ext2_read_desc {
	uint32_t first;
	uint32_t last;
	void *ptr_w;
	int visc;
};

static void
ext2_block_flush(struct ext2_read_desc *rd)
{
	int len = (rd->last - rd->first + 1) * ext2->block_size;

	ext2->lread(ext2->beg + rd->first *
			ext2->block_size / EXT2_DEVICE_BLKLEN,
			len / EXT2_DEVICE_BLKLEN,
			rd->ptr_w);

	rd->ptr_w = (uint8_t*) rd->ptr_w + len;
}

static int
ext2_block_read(struct ext2_read_desc *rd, uint32_t blkn)
{
	if (rd->first) {
		if (blkn == (rd->last + 1))
			rd->last = blkn;
		else {
			ext2_block_flush(rd);
			rd->first = rd->last = blkn;
		}
	}
	else
		rd->first = rd->last = blkn;

	if (++rd->visc >= 64) {
		serial_putc('.');
		rd->visc = 0;
	}

	return 0;
}

/* Read data blocks
 * FIXME: Overwrite is possible
 * FIXME: Total length limited: 12 + <blocksize> / 4 + (<blocksize> / 4) ^ 2 (blocks)
 * */
static uint32_t
ext2_read_data_direct(struct ext2_read_desc *rd, uint32_t *tab, uint32_t len)
{
	uint32_t i;

	for (i = 0; i < len; ++i) {
		ext2_block_read(rd, tab[i]);
	}

	return 0;
}

static uint32_t
ext2_read_data_indirect(struct ext2_read_desc *rd, uint32_t iblk, uint32_t len)
{
	uint32_t sz;

	sz = ext2->block_size / 4;

	ext2->lread(ext2->beg + iblk *
			ext2->block_size / EXT2_DEVICE_BLKLEN,
			ext2->block_size / EXT2_DEVICE_BLKLEN,
			ext2->intab);

	if (len < sz)
		sz = len;

	ext2_read_data_direct(rd, ext2->intab, sz);

	return sz;
}

static uint32_t
ext2_read_data_dindirect(struct ext2_read_desc *rd, uint32_t diblk, uint32_t len)
{
	uint32_t sz, sz2, i;
	uint32_t total;

	sz2 = ext2->block_size / 4;
	total = 0;

	ext2->lread(ext2->beg + diblk *
			ext2->block_size / EXT2_DEVICE_BLKLEN,
			ext2->block_size / EXT2_DEVICE_BLKLEN,
			ext2->dintab);

	for (i = 0; i < sz2; ++i) {

		sz = ext2_read_data_indirect(rd, ext2->dintab[i], len);

		len -= sz;
		total += sz;

		if (len == 0)
			break;
	}

	return total;
}

static int
ext2_read_data(struct ext2_inode *inode, void *ptr)
{
	struct ext2_read_desc rd;
	uint32_t size = inode->i_size;
	uint32_t blks, part;
	uint32_t sz;

	rd.ptr_w = ptr;
	rd.first = 0;
	rd.last = 0;
	rd.visc = 0;

	__u32divmod(size, ext2->block_size, &blks, &part);

	if (part)
		++blks;

	/* 12 direct blocks */
	sz = (blks < EXT2_NDIR_BLOCKS) ? blks : EXT2_NDIR_BLOCKS;
	ext2_read_data_direct(&rd, inode->i_block, sz);

	blks -= sz;
	if (blks == 0)
		goto __erd_out;

	/* indirect blocks */
	sz = ext2_read_data_indirect(&rd, inode->i_block[EXT2_IND_BLOCK], blks);

	blks -= sz;
	if (blks == 0)
		goto __erd_out;

	/* double indirect blocks */
	sz = ext2_read_data_dindirect(&rd, inode->i_block[EXT2_DIND_BLOCK], blks);

	blks -= sz;
	if (blks == 0)
		goto __erd_out;

	/* triple indirect blocks */
	//FIXME: Not implemented

__erd_out:
	ext2_block_flush(&rd);

	return 0;
}

/* Initialize ext2 structures
 * */
int ext2_mount(void *read_op, int bp)
{
	uint8_t *block;

	ext2_mem_ptr = (uint8_t*) (SDRAM_BASE_ADDRESS + SDRAM_SIZE - (1 << 20));

	block = ext2_mem_alloc(4096);
	ext2 = ext2_mem_alloc(sizeof(struct ext2_mount_data));
	ext2->block = block;
	ext2->lread = (int (*) (uint32_t, int, void*)) read_op;

	if (ext2_get_partition(bp, &ext2->beg) < 0)
		return -1;

	ext2->lread(ext2->beg + 2, 2, block);
	memcopy(&ext2->super_block, block, sizeof(struct ext2_super_block));

	if (ext2->super_block.s_magic != EXT2_SUPER_MAGIC) {
		serial_printk(EXT2 "Invalid superblock \n");
		return -1;
	}

	ext2->block_size = (1024 << ext2->super_block.s_log_block_size);
	serial_printk(EXT2 "Block size: %i \n", ext2->block_size);

	ext2->gdesc_table = ext2_mem_alloc(ext2->block_size);
	ext2->lread(ext2->beg + (ext2->super_block.s_first_data_block + 1) *
			ext2->block_size / EXT2_DEVICE_BLKLEN,
			ext2->block_size / EXT2_DEVICE_BLKLEN, ext2->gdesc_table);

	ext2_read_inode(EXT2_ROOT_INO, &ext2->root_inode);

	ext2->intab = ext2_mem_alloc(ext2->block_size);
	ext2->dintab = ext2_mem_alloc(ext2->block_size);

	return 0;
}

/* Read file from boot partition, return file size
 * */
int ext2_read(const char *name, void *ptr)
{
	struct ext2_inode *inode;
	const char *p = name;
	char part[16];		//FIXME: 15 chars file name limit
	char *w;

	serial_printk(EXT2 "Reading file \"%s\" ", name);

	inode = &ext2->curr_inode;
	memcopy(inode, &ext2->root_inode, sizeof(struct ext2_inode));

	if (*p == '/')
		++p;

	for (;;) {

		w = part;
		while ((*p) && (*p != '/')) {
			*w++ = *p++;
		}

		*w++ = 0;

		if (ext2_dir_follow(part, inode) < 0) {
			serial_printk(EXT2 "File \"%s\" not found \n", name);
			return -1;
		}

		if (*p == 0) {
			if (inode->i_mode & 0x8000)
				break;

			serial_printk(EXT2 "End point \"%s\" is not regular file \n", name);
			return -1;
		}
	}

	if (ext2_read_data(inode, ptr) < 0) {
		return -1;
	}

	serial_printk(" [%i] \n", inode->i_size);

	return inode->i_size;
}

