#include <stdint.h>

#include <mci_sd.h>
#include <AT91RM9200.h>
#include <config.h>
#include <serial.h>
#include <utils.h>

#define MCI_SD				"MCI_SD: "

#define MCI_MAX_BLOCKS			511

#define MCI_FLAG_RESP_48		0x00000040
#define MCI_FLAG_RESP_136		0x00000080
#define MCI_FLAG_OPEN_DRAIN		0x00000800
#define MCI_FLAG_MAX_LATENCY		0x00001000
#define MCI_FLAG_TRANSFER_START		0x00010000
#define MCI_FLAG_TRANSFER_STOP		0x00020000
#define MCI_FLAG_TRANSFER_READ		0x00040000
#define MCI_FLAG_TRANSFER_WRITE		0x00000000
#define MCI_FLAG_TRANSFER_SINGLE	0x00000000
#define MCI_FLAG_TRANSFER_MULTIPLE	0x00080000

struct mci_command {
	uint32_t code;
	uint32_t arg;
	uint32_t flags;
	uint32_t resp[4];
};

static uint16_t mci_sd_rca;

static int
mci_request(struct mci_command *cmd)
{
	uint32_t sreg;

	AT91C_BASE_MCI->MCI_ARGR = cmd->arg;
	AT91C_BASE_MCI->MCI_CMDR =
		(cmd->code & 0x0000003f) |
		(cmd->flags & 0x00ffffff);

	sreg = AT91C_BASE_MCI->MCI_SR;

	while (!(sreg & AT91C_MCI_CMDRDY))
		sreg = AT91C_BASE_MCI->MCI_SR;

	/* Hide fake CRC errors
	 * */
	if (cmd->code == 41)
		sreg &= ~(0x00040000);

	if (sreg & 0xc07f0000)
		serial_printk(MCI_SD "Error: MCI_SR = %x \n", sreg);

	cmd->resp[0] = AT91C_BASE_MCI->MCI_RSPR[0];
	cmd->resp[1] = AT91C_BASE_MCI->MCI_RSPR[1];
	cmd->resp[2] = AT91C_BASE_MCI->MCI_RSPR[2];
	cmd->resp[3] = AT91C_BASE_MCI->MCI_RSPR[3];

	return 0;
}

/* Only for AT91RM9200 (Errata 41.4.10)
 * */
static void
mci_reset()
{
	AT91C_BASE_MCI->MCI_CR = AT91C_MCI_MCIDIS | (1 << 7);
	AT91C_BASE_MCI->MCI_MR = AT91C_MCI_PDCMODE | ((MCI_SD_BLKLEN / 4) << 18) | (3 << 8) | (1 << 0);
	AT91C_BASE_MCI->MCI_DTOR = (0x04 << 4) | (0x04 << 0);
	AT91C_BASE_MCI->MCI_SDCR = 0;
	AT91C_BASE_MCI->MCI_CR = AT91C_MCI_MCIEN;
}

static int
mci_read_multiple(uint32_t blkn, int n, void *ptr)
{
	uint32_t wlen = n * (MCI_SD_BLKLEN / 4);
	uint32_t *ptr_s = ptr;
	uint32_t *ptr_fin = ptr_s + wlen;
	struct mci_command cmd;

	AT91C_BASE_MCI->MCI_PTCR = AT91C_PDC_TXTDIS | AT91C_PDC_RXTDIS;

	AT91C_BASE_MCI->MCI_TPR = 0;
	AT91C_BASE_MCI->MCI_TCR = 0;
	AT91C_BASE_MCI->MCI_RPR = (uint32_t) ptr;
	AT91C_BASE_MCI->MCI_RCR = wlen;

	AT91C_BASE_MCI->MCI_PTCR = AT91C_PDC_RXTEN;

	cmd.code = 18;
	cmd.arg = blkn * MCI_SD_BLKLEN;
	cmd.flags = MCI_FLAG_MAX_LATENCY | MCI_FLAG_RESP_48 |
		MCI_FLAG_TRANSFER_START | MCI_FLAG_TRANSFER_READ | MCI_FLAG_TRANSFER_MULTIPLE;

	mci_request(&cmd);

	while (!(AT91C_BASE_MCI->MCI_SR & AT91C_MCI_ENDRX)) ;

	AT91C_BASE_MCI->MCI_PTCR = AT91C_PDC_TXTDIS | AT91C_PDC_RXTDIS;

	cmd.code = 12;
	cmd.arg = 0;
	cmd.flags = MCI_FLAG_MAX_LATENCY | MCI_FLAG_RESP_48 | 
		MCI_FLAG_TRANSFER_STOP;

	mci_request(&cmd);

	mci_reset();

	/* Only for AT91RM9200 (Errata 41.4.1)
	 * */
	while (ptr_s != ptr_fin) {
		*ptr_s = swab32(*ptr_s);
		++ptr_s;
	}

	return 0;
}

int mci_sd_init()
{
	struct mci_command icmd;
	uint32_t ocr;
	uint32_t tmp32;

	AT91C_BASE_MCI->MCI_IDR = 0xffffffff;

	tmp32 = (AT91C_PA27_MCCK | AT91C_PA28_MCCDA | AT91C_PA29_MCDA0);
	AT91C_BASE_PIOA->PIO_ASR = tmp32;
	AT91C_BASE_PIOA->PIO_PDR = tmp32;

	AT91C_BASE_PMC->PMC_PCER = (1 << AT91C_ID_MCI);

	AT91C_BASE_MCI->MCI_CR = AT91C_MCI_PWSDIS | (1 << 7);
	AT91C_BASE_MCI->MCI_MR = AT91C_MCI_PDCMODE | ((MCI_SD_BLKLEN / 4) << 18) | (3 << 8) | (80 << 0);
	AT91C_BASE_MCI->MCI_DTOR = (0x04 << 4) | (0x04 << 0);
	AT91C_BASE_MCI->MCI_SDCR = 0;
	AT91C_BASE_MCI->MCI_CR = AT91C_MCI_MCIEN;

	/* Reset SD
	 * */
	icmd.code = 0;
	icmd.arg = 0;
	icmd.flags = MCI_FLAG_MAX_LATENCY;

	mci_request(&icmd);

	nops(1000);

	/* Voltage range: 3.2 - 3.4
	 * */
	ocr = (1 << 20) | (1 << 21);

	/* Wait for power up
	 * */
	tmp32 = 1000;

	while (tmp32) {

		icmd.code = 55;
		icmd.arg = 0;
		icmd.flags = MCI_FLAG_MAX_LATENCY | MCI_FLAG_RESP_48;

		mci_request(&icmd);

		icmd.code = 41;
		icmd.arg = ocr;
		icmd.flags = MCI_FLAG_MAX_LATENCY | MCI_FLAG_RESP_48;

		mci_request(&icmd);

		if (icmd.resp[0] & (1 << 31))
			break;

		nops(1000);
		--tmp32;
	}

	if (tmp32 == 0) {
		serial_printk(MCI_SD "Card not responding \n");
		return -1;
	}

	/* Go to transfer mode
	 * */
	icmd.code = 2;
	icmd.arg = 0;
	icmd.flags = MCI_FLAG_MAX_LATENCY | MCI_FLAG_RESP_136;

	mci_request(&icmd);

	serial_printk(MCI_SD "CID: [%x %x %x %x] \n",
			icmd.resp[0], icmd.resp[1], icmd.resp[2], icmd.resp[3]);

	icmd.code = 3;
	icmd.arg = 0;
	icmd.flags = MCI_FLAG_MAX_LATENCY | MCI_FLAG_RESP_48;

	mci_request(&icmd);

	mci_sd_rca = ((icmd.resp[0] & 0xffff0000) >> 16);

	serial_printk(MCI_SD "RCA: [%h] \n", mci_sd_rca);

	/* Get CSD
	 * */
	icmd.code = 9;
	icmd.arg = ((uint32_t) mci_sd_rca) << 16;
	icmd.flags = MCI_FLAG_MAX_LATENCY | MCI_FLAG_RESP_136;

	mci_request(&icmd);

	serial_printk(MCI_SD "CSD: [%x %x %x %x] \n",
			icmd.resp[0], icmd.resp[1], icmd.resp[2], icmd.resp[3]);

	/* Go to transfer state
	 * */
	icmd.code = 7;
	icmd.arg = ((uint32_t) mci_sd_rca) << 16;
	icmd.flags = MCI_FLAG_MAX_LATENCY | MCI_FLAG_RESP_48;

	mci_request(&icmd);

	serial_printk(MCI_SD "Card Status: [%x] \n", icmd.resp[0]);

	/* Clock up and reset
	 * */
	mci_reset();

	/* Set block length
	 * */
	icmd.code = 16;
	icmd.arg = MCI_SD_BLKLEN;
	icmd.flags = MCI_FLAG_MAX_LATENCY | MCI_FLAG_RESP_48;

	mci_request(&icmd);

	return 0;
}

int mci_sd_read(uint32_t blkn, int n, void *ptr)
{
	uint8_t *ptr_w = ptr;
	int gn;

	while (n > 0) {

		gn = (n < MCI_MAX_BLOCKS) ? n : MCI_MAX_BLOCKS;
		mci_read_multiple(blkn, gn, ptr_w);

		blkn += gn;
		n -= gn;
		ptr_w += gn * MCI_SD_BLKLEN;
	}

	return 0;
}

